(ns puppetlabs.classifier.rules
  (:require [clojure.tools.logging :as log]
            [clojure.walk :as walk]
            [fast-zip.core :as z]
            [fast-zip.visit :as zv]
            [schema.core :as sc]
            [slingshot.slingshot :refer [throw+]]
            [puppetlabs.kitchensink.core :refer [parse-number]]
            [puppetlabs.classifier.schema :refer [ExplainedCondition MeaningfulRuleCondition Rule
                                                  RuleCondition]])
  (:import java.util.regex.Pattern
           java.util.UUID))

(def always-matches ["~" "name" ".*"])

(def numeric-operator? (partial contains? #{"<" ">" "<=" ">="}))

(defn- lookup-field
  [m ks]
  (let [get* (fn [m k] (get m (keyword k) (get m (str k))))]
    (if (sequential? ks)
      (reduce get* m ks)
      (get* m ks))))

(defn- safer-re-find
  [target-value node-value]
  (if (string? node-value)
    (re-find (Pattern/compile target-value) node-value)))

(sc/defn match? :- Boolean
  [rule-condition :- RuleCondition, node]
  (let [[operator & args] rule-condition]
    (case operator
      "not" (not (match? (first args) node))
      "and" (every? identity (map #(match? % node) args))
      "or" (boolean (some identity (map #(match? % node) args)))
      (let [[field target] args
            node-value (let [v (lookup-field node field)]
                         (cond
                           (and (numeric-operator? operator) (string? v)) (parse-number v)
                           (or (true? v) (false? v)) (str v)
                           (and (= operator "=") (number? v)) (str v)
                           :else v))
            target-value (if (and (numeric-operator? operator) (string? target))
                           (parse-number target)
                           target)]
        (if (and node-value target-value)
          (case operator
            "=" (= node-value target-value)
            "~" (boolean (safer-re-find target-value node-value))
            "<" (< node-value target-value)
            "<=" (<= node-value target-value)
            ">" (> node-value target-value)
            ">=" (>= node-value target-value)
            false)
          false)))))

(sc/defn apply-rule :- (sc/maybe UUID)
  "Apply a single rule to a node"
  [node rule :- Rule]
  (try (if (match? (:when rule) node)
         (:group-id rule))
    (catch Exception e
      ;; It's usually bad form to swallow all exceptions like this but if any
      ;; exception makes it out of here, then classification for all nodes will
      ;; fail. So, in this case, we'll swallow them, log the exception, and
      ;; return nil
      (log/warn e
                "encountered an exception when trying to match rule" (pr-str rule)
                "of group" (:group-id rule)
                "against the node named" (:name node))
      nil)))

(sc/defn explain-rule* :- ExplainedCondition
  [node, condition :- RuleCondition]
  (let [recurse (partial explain-rule* node)
        op (first condition)]
    (case op
      "not" {:value (match? condition node)
             :form ["not" (recurse (last condition))]}

      ("and" "or") {:value (match? condition node)
                    :form (vec (cons op (map recurse (rest condition))))}

      ("=" "~" "<" "<=" ">" ">=") {:value (match? condition node)
                                   :form (let [[op field target] condition
                                               node-value {:path field
                                                           :value (lookup-field node field)}]
                                           [op node-value target])})))

(sc/defn explain-rule :- ExplainedCondition
  [rule :- RuleCondition, node]
  (explain-rule* node rule))

(sc/defn rectify-meaningless-rule :- MeaningfulRuleCondition
  "Turn a meaningless rule (i.e. one with a numeric comparison against
  a non-numeric target) into a similarly evaluating rule that explains what the
  old, meaningless rule got wrong."
  [rule :- RuleCondition]
  (let [xform-meaningless (fn [form]
                            (cond
                              (not (sequential? form)) form
                              (not (contains? #{"<" ">" "<=" ">="} (nth form 0))) form
                              :else (let [[op val-path target] form]
                                      (if (parse-number target)
                                        form
                                        ["="
                                         val-path
                                         (str "this rule clause used to be a " op " comparison"
                                              " against a non-numeric target value of"
                                              " " (pr-str target))]))))]
    (walk/prewalk xform-meaningless rule)))

(defn rule-branch?
  "Rule branches are those rule clauses that can contain sub-clauses, namely
  the \"and\", \"or\", and \"not\" operators."
  [clause]
  (contains? #{"and" "or" "not"} (first clause)))

(def rule-zipper
  (partial z/zipper
           rule-branch?
           rest
           (fn [[op] sub-clauses] (apply vector op sub-clauses))))

(sc/defn facts-of
  "Returns a collection of all the fact paths referenced by the rule."
  [rule :- RuleCondition]
  (let [factfinder (zv/visitor :pre [rule facts]
                               {:state (if-not (rule-branch? rule)
                                         (conj facts (nth rule 1))
                                         facts)})]
    (:state (zv/visit (rule-zipper rule) [] [factfinder]))))

(sc/defn comparisons-of
  "Returns a vector of [<operator>, <fact-path>] for every comparison clause in
  the given rule."
  [rule]
  (let [compfinder (zv/visitor :pre [rule comps]
                               {:state (if-not (rule-branch? rule)
                                         (conj comps (vec (take 2 rule)))
                                         comps)})]
    (:state (zv/visit (rule-zipper rule) [] [compfinder]))))

(defn rand-value-for-operator
  "Returns a random value to use as the node value operand in a comparison rule
  clause. For `=` and `~` operators, it is a random string of printable ASCII
  characters; for the numeric comparison operators, it is the string
  representation of a random integer in the half-open range [0, 100)."
  [op]
  (let [rand-str (apply str (map (comp char (partial + 32))
                                 (repeatedly 15 #(rand-int 91))))]
    (case op
      ("=" "~") rand-str
      ("<" ">" "<=" ">=") (str (rand-int 100)))))

(sc/defn stubbed-node
  "Returns a node map stubbed out with random values for every fact path used by
  the rule."
  [rule]
  (reduce (fn [node [op fact]]
            (let [fact-path (if (coll? fact) fact [fact])]
              (assoc-in node fact-path (rand-value-for-operator op))))
          {"name" "test"}
          (comparisons-of rule)))

(defn valid-rule?
  "Returns true iff the rule is well-formed, meaningful, and can be applied to
  a node that has random values for all the facts referenced in the rule without
  throwing any exceptions."
  [rule]
  (if-not (nil? (sc/check MeaningfulRuleCondition rule))
    false
    (try
      (match? rule (stubbed-node rule))
      true
      (catch Exception _
        false))))
