;; Portions of this file use code from Pocheshiro (c) 2013 INVENTI
;; under the terms of the MIT license.

(ns puppetlabs.rbac.services.http.middleware
  (:require [puppetlabs.rbac.shiro-utils :as shiro]
            [slingshot.slingshot :refer [throw+ try+]]
            [puppetlabs.kitchensink.core :refer [cn-whitelist->authorizer cn-for-cert]]
            [puppetlabs.trapperkeeper.core :refer [defservice]]
            [puppetlabs.trapperkeeper.services :refer [service-context]]) 
  (:import [org.apache.shiro.web.servlet ShiroHttpServletRequest]
           [org.apache.shiro.subject.support SubjectCallable]
           [org.apache.shiro.subject ExecutionException]
           [org.apache.shiro.web.subject WebSubject$Builder]))

(defn- non-session-encoding-response
  "By default, HttpServletResponseWrapper adds JSESSIONID to URLs as a
  way of doing non-cookie session management. We do not need
  this. This function wraps a servlet response with a proxy that
  leaves the URLs unaltered.

  Adapted from Pocheshiro."
  [servlet-response]
  (proxy [javax.servlet.http.HttpServletResponseWrapper] [servlet-response]
    (encodeURL [url] url)
    (encodeUrl [url] url)
    (encodeRedirectURL [url] url)
    (encodeRedirectUrl [url] url)))

(defn ssl?
  "Checks whether a given ring request is using ssl by checking its
  scheme and ssl-client-cert keys."
  [request]
  (and (= :https (:scheme request))
       (some? (:ssl-client-cert request))))

(defn augment-request
  "Given a Shiro security manager and a ring request map, this
  performs several operations on the map and returns an updated
  version:

  * adds the attached SSL cert's CN, if there is one
  * adds a servlet-response that does not encode any session info in redirect URLs (the servlet default)
  * adds a shiro request
  * adds the passed security manager"
  [sec-mgr {:keys [servlet-request
                   servlet-context
                   servlet-response] :as request} & [subject]]
  (let [servlet-response (non-session-encoding-response servlet-response)
        shiro-request    (ShiroHttpServletRequest. servlet-request servlet-context true)
        ssl-client-cn    (when (ssl? request) (cn-for-cert (:ssl-client-cert request)))]

    (assoc request
      :shiro-request    shiro-request
      :ssl-client-cn    ssl-client-cn
      :servlet-response servlet-response
      :sec-mgr          sec-mgr)))

(defn- execute-handler
  "Given a handler and a ring request augmented by `augment-request`,
  this function executes the handler for the request while binding
  Shiro's current subject to the shiro subject found in the augmented
  request. This pinning persists across threads, making Shiro's magic
  getSubject function threadsafe and deterministic for the duration of
  the request fulfillment.

  Any exception thrown during the course of request fulfillment is
  caught and re-thrown upwards."
  [handler {:keys [shiro-subject] :as request}]
  (try
    (.call (SubjectCallable. shiro-subject #(handler request)))
    (catch ExecutionException e (throw (.getCause e)))))

(defn format-redirect
  "Given a request, prepare the redirect URL by potentially appending
  a query string."
  [request]
  (if (empty? (:query-string request))
    (:uri request)
    (format "%s?%s" (:uri request) (:query-string request))))

(defn throw-unauthenticated! [request]
  (throw+ {:kind :puppetlabs.rbac/user-unauthenticated
           :msg "Route requires authentication"
           :redirect (format-redirect request)}))

(defn throw-revoked! []
  (throw+ {:kind :puppetlabs.rbac/user-revoked
           :msg  "User is revoked and may not access protected routes."}))

(defn build-web-subject
  "Given a request run through (augment-request), produce a Shiro
  WebSubject suitable for authentication and authorization checks."
  [{:keys [sec-mgr shiro-request servlet-response]}]
  (.buildWebSubject (WebSubject$Builder. sec-mgr shiro-request servlet-response)))

(defn handle-cert-access
  "Given the TK context, a Shiro SecurityManager instance, a ring
  handler and a request passed through (augment-request), checks to
  see if:

  1. The ssl cert encrypting this request matches the cert whitelist
  2. The API user (who is used by default for cert-authenticated requests) has not been revoked.

  If both checks succeed, the handler is called.

  For now, this function uses the admin user; in PE-5668 it will use a
  specific API user"
  [{:keys [cert-whitelisted? is-revoked?]} sec-mgr handler request]
  ;; TODO use api_user once PE-5668 is done
  (let [shiro-subject (shiro/login->subject sec-mgr "admin")]

    (when-not (cert-whitelisted? request)
      (throw-unauthenticated! request))

    (when (is-revoked? shiro-subject)
      (throw-revoked!))

    (execute-handler handler (assoc request :shiro-subject shiro-subject))))

(defn handle-web-access
  "Given the TK context, a Shiro SecurityManager instance, a ring
  handler and a request passed through (augment-request), checks to
  see if:

  1. The WebSubject associated with the request's session cookie is authenticated
  2. The WebSubject associated with the request's session cookie is revoked

  If both checks succeed, the handler is called."
  [{:keys [is-revoked?]} sec-mgr handler request]
  (let [shiro-subject (build-web-subject request)]

    (when-not (shiro/authenticated? shiro-subject)
      (throw-unauthenticated! request))

    (when (is-revoked? shiro-subject)
      (throw-revoked!))

    (execute-handler handler (assoc request :shiro-subject shiro-subject))))

(defn do-wrap-authenticated
  "Given an is-revoked? that accepts a Shiro subject, a function for
  retrieving a Shiro security manager and a handler (presumably
  created with compojure or similar), this function wraps the handler
  with checks for authentication status and revoked status of the
  accessing subject.

  First, the subject is checked for authentication; if it is
  unauthenticated, handle-unauthenticated is called. Next, the
  authenticated subject is checked to see if it has been revoked; if
  it is, handle-revoked is called. Finally, the handler is executed
  with a pinned shiro subject (making all calls to getSubject
  thread-safe and deterministic). "
  [{:keys [whitelist-certs?] :as ctx} security-manager handler]
  (fn [request]
    (let [sec-mgr (security-manager)
          request' (augment-request sec-mgr request)]
      (if (or (not (ssl? request'))
              (not whitelist-certs?))
        ;; No SSL cert or whitelist means all we can do is look for a
        ;; JSESSIONID.
        (handle-web-access ctx sec-mgr handler request')
        (try+
         ;; Now, it is unclear if the request is API or web. Try
         ;; web first, letting a revoked exception bubble up, but retry
         ;; with cert whitelisting if unauthenticated is thrown.
         (handle-web-access ctx sec-mgr handler request')
         (catch [:kind :puppetlabs.rbac/user-unauthenticated] e
           (handle-cert-access ctx sec-mgr handler request')))))))

(defn do-wrap-unauthenticated [security-manager handler]
  (fn [request]
    (let [request' (augment-request (security-manager) request)]
      (execute-handler handler (assoc request' :shiro-subject (build-web-subject request'))))))

(defprotocol RbacAuthnMiddleware
  "We expose our authentication middleware as a TK service so it can
  access the security manager, config and database state managed by TK
  for our application."
  (wrap-unauthenticated [this handler])
  (wrap-authenticated [this handler]))

(defservice rbac-authn-middleware
  RbacAuthnMiddleware
  [[:RbacService security-manager get-subject]
   [:ConfigService get-in-config]]
  (init [this context]
        (let [cert-whitelist (get-in-config [:global :certificate-whitelist])
              whitelist-certs? (not (empty? cert-whitelist))
              is-revoked? #(->> (.getPrincipal %)
                                (get-subject :login)
                                :is_revoked)]
          (assoc context
            :whitelist-certs? whitelist-certs?
            :cert-whitelisted? (if whitelist-certs?
                                 (cn-whitelist->authorizer cert-whitelist)
                                 (constantly false)) ;; should not be called if no whitelist
            :is-revoked? is-revoked?)))
  (wrap-unauthenticated [this handler]
                        (do-wrap-unauthenticated security-manager handler))
  (wrap-authenticated [this handler]
                       (do-wrap-authenticated (service-context this)
                                              security-manager
                                              handler)))
